/* This file is an amalgamation of several source files.
   Please look for BEGIN FILE,END FILE and CONTINUE FILE
   strings to browse contents of individual files contained
   here. */
/* BEGIN FILE main.c */
/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <cinarus@yahoo.com> wrote this file. As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.
 *
 * I don't guarantee that this program will work at all or that it won't
 *  delete all data on your disk. Use it at your own risk.
 * ----------------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
/* BEGIN FILE simple.c */
void replace_control_chars(uint8_t *data, size_t *Rlen)
{
   size_t i,len;
   len= *Rlen;
   for(i=0;i<len;i++)
   {
      switch(data[i])
      {
      case 0x7f:
      case 0x00: case 0x01: case 0x02: case 0x03:
      case 0x04: case 0x05: case 0x06: case 0x07:
      case 0x08: case 0x09:            case 0x0b:
      case 0x0c:            case 0x0e: case 0x0f:
      case 0x10: case 0x11: case 0x12: case 0x13:
      case 0x14: case 0x15: case 0x16: case 0x17:
      case 0x18: case 0x19: case 0x1a: case 0x1b:
      case 0x1c: case 0x1d: case 0x1e: case 0x1f:
         data[i]= ' ';
         break;
      }
   }
   if (len && data[len-1]=='\n') { data[len-1]= 0; len--; }
   if (len && data[len-1]=='\r') { data[len-1]= 0; len--; }
   *Rlen= len;
}
int remove_line_continuations(uint8_t *data, size_t len)
{
   int R;
   size_t i,j;
   R= 0;
   for(i=0;i<len;i++)
   {
     if (data[i]!='\\') continue;
     for(j=i+1;j<len;j++) if (data[j]!=' ') break;
     if (j<len && data[j]!='\n') continue;
     if (j==len) j--;
     for(;i<j+1;i++) { data[i]= 0; R++; }
     i--;
   }
   return R;
}
/* END FILE simple.c */
/* CONTINUE FILE main.c */
/* BEGIN FILE ifile.c */
/* BEGIN FILE ifile.h */
#ifndef ifile_h_included
#define ifile_h_included
typedef struct ifile ifile_t;
int ifile_open(ifile_t **Rf,char *name);
int ifile_get_line(ifile_t *f,uint8_t **buffer, size_t *len,int nullterm);
void ifile_close(ifile_t *f);
int ifile_line_number(ifile_t *f);
#endif
/* END FILE ifile.h */
/* CONTINUE FILE ifile.c */
#define IFILE_BLOCK_SIZE 128
typedef struct ifileblock {
  uint8_t data[IFILE_BLOCK_SIZE];
  size_t len;
  struct ifileblock *next;
} ifileblock_t;
struct ifile {
  FILE *handle;
  ifileblock_t *blocks,*last;
  int eof, lineno;
  char *name, *error;
};
int ifile_open(ifile_t **Rf,char *name)
{
  ifile_t *f;
  char buf[2000];
  int R;
  R= 0;
  f= malloc(sizeof(*f));
  f->handle= fopen(name, "r");
  if (!f->handle)  {
     snprintf(buf, sizeof(buf), "can't open %s: %s",name, strerror(errno));
     f->error= strdup(buf);
     R= 1;
  }
  f->eof= 0;
  f->lineno= 0;
  f->name= strdup(name);
  f->blocks= f->last= NULL;
  f->error= NULL;
  *Rf= f;
  return R;
}
static int ifile_read_block(ifile_t *f)
{
  ifileblock_t *B;
  size_t R;
  if (f->eof) return 1;
  B= malloc(sizeof(*B));
  R= fread(B->data, 1, sizeof(B->data), f->handle);
  if (R<sizeof(B->data))
  {
     if (R==0) free(B);
     f->eof= 1;
     if (ferror(f->handle))
     {
       char buf[2000];
       snprintf(buf,sizeof(buf), "read error from %s: %s",
                          f->name, strerror(errno));
       f->error= strdup(buf);
     }
     if (R==0) return 1;
  }
  B->len= R;
  B->next= NULL;
  if (!f->blocks) { f->blocks= f->last= B; }
  else { f->last->next= B; f->last= B; }
  return 0;
}
static void ifile_collect_blocks
    (ifile_t *f, ifileblock_t *E,size_t endi,
     uint8_t **Rbuffer, size_t *Rlen, int nullterm)
{
   ifileblock_t *B,*N;
   uint8_t *buffer;
   size_t size;
   size= 0;
   if (endi==E->len-1) { E=NULL; endi= -1; }
   for(B=f->blocks;B!=E;B=B->next) size+= B->len;
   size+= endi+1;
   *Rlen= size;
   buffer= malloc(size + nullterm);
   size= 0;
   for(B=f->blocks;B!=E;B=N)
   {
     N= B->next;
     memcpy(buffer+size, B->data, B->len);
     size+= B->len;
     free(B);
   }
   if (E) {
     memcpy(buffer+size, E->data, endi+1);
     memmove(E->data, E->data+endi+1, E->len-endi-1);
     E->len-= endi+1;
     size+= endi+1;
   }
   f->blocks= E;
   *Rbuffer= buffer;
   if (nullterm) { int i; for(i=0;i<nullterm;i++) buffer[size+i]= 0; }
}
int ifile_get_line(ifile_t *f,uint8_t **buffer, size_t *len,int nullterm)
{
  ifileblock_t *L;
  int i;
  if (!f->blocks && ifile_read_block(f)) return 1;
again:
  i= 0;
  L= f->last;
next:
  if (i==L->len)
  {
    if (ifile_read_block(f)) goto done;
    goto again;
  }
  if (L->data[i]=='\n') goto done;
  if (L->data[i]=='\r')
  {
     if (i==L->len-1) {
        if (ifile_read_block(f)) goto done;
        if (L->next->data[0]=='\n') { i= 0; L= L->next; }
     } else {
        if (L->data[i+1]=='\n') i++;
     }
     goto done;
  }
  i++;
  goto next;
done:
  ifile_collect_blocks(f, L,i, buffer, len,nullterm);
  f->lineno++;
  return 0;
}
void ifile_close(ifile_t *f)
{
   ifileblock_t *B,*N;
   if (f->handle) fclose(f->handle);
   for(B=f->blocks;B;B=N) { N= B->next; free(B); }
   if (f->error) free(f->error);
   free(f->name);
   free(f);
}
int ifile_line_number(ifile_t *f) { return f->lineno; }
char *ifile_error(ifile_t *f) { return f->error; }
/* END FILE ifile.c */
/* CONTINUE FILE main.c */
/* BEGIN FILE obuffer.c */
#include <stdarg.h>
#define OBUFFER_INITIAL_SIZE 128
typedef struct {
  uint8_t *data;
  size_t size, len, head;
} obuffer_t;
obuffer_t *obuffer_new()
{
   obuffer_t *V;
   V= malloc(sizeof(*V));
   V->size= OBUFFER_INITIAL_SIZE;
   V->len= V->head= 0;
   V->data= malloc(V->size);
   return V;
}
void obuffer_put(obuffer_t *V, const char *fmt, ...)
{
   va_list ap;
   int i;
   size_t size;
   size_t copy_sz; void *copy_p;
   int intv; double douv; void *ptrv;
   int copy;
   copy= 0;
again:
   size= 0;
   va_start(ap,fmt);
   for(i=0;fmt[i];i++)
   {
    switch(fmt[i])
    {
    case 'i': intv= va_arg(ap,int); copy_sz= sizeof(int);
                                    copy_p= &intv; break;
    case 'd': douv= va_arg(ap, double); copy_sz= sizeof(double);
                                        copy_p= &douv; break;
    case 'p': ptrv= va_arg(ap, void*); copy_sz= sizeof(void*);
                                       copy_p= &ptrv; break;
    case 'b': copy_sz= va_arg(ap, size_t); copy_p= va_arg(ap, void*); break;
    }
    if (copy) { memcpy(V->data+V->len, copy_p, copy_sz); V->len+= copy_sz; }
    else size+= copy_sz;
   }
   va_end(ap);
   if (copy) return ;
   copy++;
   if (V->len+size>V->size)
   {
     V->size+= V->size/2 + size;
     V->data= realloc(V->data, V->size);
   }
   goto again;
}
int obuffer_empty(obuffer_t *o)
{
   return o->head==o->len;
}
void obuffer_read(obuffer_t *o, void *dst, size_t sz)
{
   memcpy(dst, o->data+o->head, sz);
   o->head+= sz;
}
double obuffer_dou(obuffer_t *o)
  { double D; obuffer_read(o,&D,sizeof(D)); return D; }
void* obuffer_ptr(obuffer_t *o)
  { void* D; obuffer_read(o,&D,sizeof(D)); return D; }
int obuffer_int(obuffer_t *o)
  { int D; obuffer_read(o,&D,sizeof(D)); return D; }
void obuffer_free(obuffer_t *o) { free(o->data); free(o); }
/* END FILE obuffer.c */
/* CONTINUE FILE main.c */
/* BEGIN FILE linebuffer.c */
void linebuffer_collect(obuffer_t *B,uint8_t **Rbuffer,size_t *Rlen)
{
   obuffer_t *L;
   uint8_t *line; size_t len;
   int size;
   uint8_t *buffer;
   int copy;
   copy= 0;
again:
   size= 0;
   B->head= 0;
   while(!obuffer_empty(B))
   {
      len= obuffer_int(B);
      line= obuffer_ptr(B);
      if (copy) {
        memcpy(buffer+size, line, len);
        free(line);
      }
      size+= len;
   }
   if (copy) { obuffer_free(B); return ; }
   copy++;
   buffer= malloc(size+1); buffer[size]= 0;
   *Rbuffer= buffer; *Rlen= size;
   goto again;
}
void linebuffer_add(obuffer_t *B, uint8_t *line, size_t len)
{
   obuffer_put(B,"ip", len, line);
}
/* END FILE linebuffer.c */
/* CONTINUE FILE main.c */
/* BEGIN FILE util.c */
void strdel(char *str, int start, int end)
{
  size_t len;
  len= strlen(str);
  if (start<0) start=len+start; if (start<0) return ;
  if (end<0)   end= len+end;    if (end<0) return ;
  if (end<start) { int t= end; end= start; start= t; }
  if (end>=len) end= len-1;
  if (start>=len) return ;
  memmove(str+start, str+end+1, len-end);
}
void remove_string(uint8_t *buffer, size_t first, size_t last, size_t *len)
{
  memmove(buffer+first, buffer+last+1, *len - last-1);
  *len-= (last-first+1);
  buffer[*len]= 0;
}
int skip_quote(uint8_t *buffer, size_t pos, size_t len)
{
  uint8_t Q;
  Q= buffer[pos++];
again:
  if (pos==len) return pos;
  if (buffer[pos]=='\\') {
      pos++; if (pos==len) return len;
      pos++; if (pos==len) return len;
      goto again;
  }
  if (buffer[pos]==Q) return pos;
  pos++;
  goto again;
}
void remove_trailing_spaces(uint8_t *buffer, size_t *Rlen)
{
  size_t len;
  len= *Rlen;
  while(len>0 && buffer[len-1]==' ') { len--; buffer[len]= 0; }
  *Rlen= len;
}
void remove_leading_spaces(uint8_t *buffer, size_t *Rlen)
{
  size_t len,i;
  len= *Rlen;
  for(i=0;i<len;i++)
      if (buffer[i]!=' ') break;
  if (i==0) return ;
  remove_string(buffer, 0, i-1, Rlen);
}
/* END FILE util.c */
/* CONTINUE FILE main.c */
/* BEGIN FILE combine.c */
int get_combined_line(ifile_t *file,
     uint8_t **Rbuffer, size_t *Rlen,int *Rlineno)
{
   obuffer_t *L;
   uint8_t *line; size_t len;
   int lineno;
   L= NULL; line= NULL; len= 0; lineno= 0;
   while(!ifile_get_line(file, &line, &len, 1))
   {
      if (!lineno) lineno= ifile_line_number(file);
      replace_control_chars(line, &len);
      remove_trailing_spaces(line, &len);
      if (!len || line[len-1]!='\\') break;
      if (!L) L= obuffer_new();
      line[len-1]= 0;
      linebuffer_add(L, line, len-1);
   }
   if (!L && !line) return 1;
   *Rlineno= lineno;
   if (!L) { *Rlen= len; *Rbuffer= line; return 0; }
   linebuffer_add(L, line, len);
   linebuffer_collect(L, Rbuffer, Rlen);
   return 0;
}
/* END FILE combine.c */
/* CONTINUE FILE main.c */
/* BEGIN FILE comment.c */
FILE *comment_out;
void write_comment(uint8_t *buffer, size_t len)
{
  if (comment_out) {
    fwrite(buffer, 1, len, comment_out);
    fprintf(comment_out,"\n");
  }
}
void remove_comments(ifile_t *file,uint8_t **Rbuffer,size_t *Rlen)
{
  size_t i,j;
  obuffer_t *L;
  int dummy;
  size_t len;
  uint8_t *buffer;
  len= *Rlen;
  buffer= *Rbuffer;
  i= 0;
again:
  for(;i<len;i++)
   if (buffer[i]=='\'' || buffer[i]=='"') i= skip_quote(buffer,len,i);
   else if (buffer[i]=='/' && buffer[i+1]=='*') goto block_comment;
   else if (buffer[i]=='/' && buffer[i+1]=='/') goto line_comment;
done:
  *Rbuffer= buffer;
  *Rlen= len;
  return ;
line_comment:
  write_comment(buffer+i, len-i);
  len= i;
  buffer[len]= 0;
  goto done;
block_comment:
  for(j=i+2;j<len;j++)
    if (buffer[j]=='*' && buffer[j+1]=='/') break;
  if (j!=len)
  {
    write_comment(buffer+i, j+2-i);
    buffer[i]= ' ';
    remove_string(buffer, i+1, j+1,&len);
    goto again;
  }
  write_comment(buffer+i, len-i);
  buffer[i]= ' '; buffer[i+1]= 0; len= i+1;
  L= obuffer_new();
  linebuffer_add(L, buffer, len);
  while(!get_combined_line(file, &buffer, &len, &dummy))
  {
    for(j=0;j<len;j++)
      if (buffer[j]=='*' && buffer[j+1]=='/') break;
    if (j==len) { write_comment(buffer, len); free(buffer); continue; }
    write_comment(buffer, j+2);
    remove_string(buffer, 0, j+1,&len);
    linebuffer_add(L, buffer,len);
    break;
  }
  linebuffer_collect(L, &buffer, &len);
  goto again;
}
/* END FILE comment.c */
/* CONTINUE FILE main.c */
/* BEGIN FILE cmdline.c */
/* Generated by re2c 0.13.5 on Thu Jun 13 02:49:02 2013 */
#line 1 "cmdline.re2c"
/* BEGIN FILE cmdline.h */
#ifndef _cmdline_h_included
#define _cmdline_h_included
#ifdef __cplusplus
extern "C" {
#endif
char** parse_cmd_line(char *inp,int *err);
#ifdef __cplusplus
}
#endif
#endif
/* END FILE cmdline.h */
/* CONTINUE FILE cmdline.c */
char** parse_cmd_line(char *inp,int *err)
{
   int N;
   char **out;
   char *begin,*cursor,*marker;
   out= 0;
   *err= -1;
restart:
   N= 0;
   cursor= inp;
next_token:
   begin= cursor;
#line 24 "<stdout>"
{
 char yych;
 yych = *cursor;
 switch (yych) {
 case 0x00:
 case '#': goto yy2;
 case 0x01:
 case 0x02:
 case 0x03:
 case 0x04:
 case 0x05:
 case 0x06:
 case 0x07:
 case 0x08:
 case '\t':
 case '\v':
 case '\f':
 case '\r':
 case 0x0E:
 case 0x0F:
 case 0x10:
 case 0x11:
 case 0x12:
 case 0x13:
 case 0x14:
 case 0x15:
 case 0x16:
 case 0x17:
 case 0x18:
 case 0x19:
 case 0x1A:
 case 0x1B:
 case 0x1C:
 case 0x1D:
 case 0x1E:
 case 0x1F:
 case ' ':
 case 0x7F: goto yy5;
 case '\n': goto yy4;
 case '"': goto yy7;
 case '\'': goto yy9;
 case '\\': goto yy12;
 default: goto yy10;
 }
yy2:
 ++cursor;
yy3:
#line 27 "cmdline.re2c"
 { goto done; }
#line 75 "<stdout>"
yy4:
 yych = *++cursor;
 goto yy3;
yy5:
 ++cursor;
 yych = *cursor;
 goto yy27;
yy6:
#line 28 "cmdline.re2c"
 { goto next_token; }
#line 86 "<stdout>"
yy7:
 yych = *(marker = ++cursor);
 switch (yych) {
 case 0x00:
 case '\n': goto yy8;
 default: goto yy22;
 }
yy8:
#line 32 "cmdline.re2c"
 { goto fail; }
#line 97 "<stdout>"
yy9:
 yych = *(marker = ++cursor);
 switch (yych) {
 case 0x00:
 case '\n': goto yy8;
 default: goto yy16;
 }
yy10:
 ++cursor;
 yych = *cursor;
 goto yy14;
yy11:
#line 31 "cmdline.re2c"
 { goto record; }
#line 112 "<stdout>"
yy12:
 yych = *++cursor;
 goto yy8;
yy13:
 ++cursor;
 yych = *cursor;
yy14:
 switch (yych) {
 case 0x00:
 case 0x01:
 case 0x02:
 case 0x03:
 case 0x04:
 case 0x05:
 case 0x06:
 case 0x07:
 case 0x08:
 case '\t':
 case '\n':
 case '\v':
 case '\f':
 case '\r':
 case 0x0E:
 case 0x0F:
 case 0x10:
 case 0x11:
 case 0x12:
 case 0x13:
 case 0x14:
 case 0x15:
 case 0x16:
 case 0x17:
 case 0x18:
 case 0x19:
 case 0x1A:
 case 0x1B:
 case 0x1C:
 case 0x1D:
 case 0x1E:
 case 0x1F:
 case ' ':
 case '"':
 case '#':
 case '\'':
 case '\\':
 case 0x7F: goto yy11;
 default: goto yy13;
 }
yy15:
 ++cursor;
 yych = *cursor;
yy16:
 switch (yych) {
 case 0x00:
 case '\n': goto yy17;
 case '\'': goto yy19;
 case '\\': goto yy18;
 default: goto yy15;
 }
yy17:
 cursor = marker;
 goto yy8;
yy18:
 ++cursor;
 yych = *cursor;
 switch (yych) {
 case 0x00:
 case '\n': goto yy17;
 default: goto yy15;
 }
yy19:
 ++cursor;
#line 30 "cmdline.re2c"
 { goto record; }
#line 187 "<stdout>"
yy21:
 ++cursor;
 yych = *cursor;
yy22:
 switch (yych) {
 case 0x00:
 case '\n': goto yy17;
 case '"': goto yy24;
 case '\\': goto yy23;
 default: goto yy21;
 }
yy23:
 ++cursor;
 yych = *cursor;
 switch (yych) {
 case 0x00:
 case '\n': goto yy17;
 default: goto yy21;
 }
yy24:
 ++cursor;
#line 29 "cmdline.re2c"
 { goto record; }
#line 211 "<stdout>"
yy26:
 ++cursor;
 yych = *cursor;
yy27:
 switch (yych) {
 case 0x01:
 case 0x02:
 case 0x03:
 case 0x04:
 case 0x05:
 case 0x06:
 case 0x07:
 case 0x08:
 case '\t':
 case '\v':
 case '\f':
 case '\r':
 case 0x0E:
 case 0x0F:
 case 0x10:
 case 0x11:
 case 0x12:
 case 0x13:
 case 0x14:
 case 0x15:
 case 0x16:
 case 0x17:
 case 0x18:
 case 0x19:
 case 0x1A:
 case 0x1B:
 case 0x1C:
 case 0x1D:
 case 0x1E:
 case 0x1F:
 case ' ':
 case 0x7F: goto yy26;
 default: goto yy6;
 }
}
#line 33 "cmdline.re2c"
done:
   if (out) { out[N]=0; return out; }
   if (N==0) return 0;
   out= malloc(sizeof(char*)*(N+1));
   goto restart;
record:
   if (out) {
      int L= cursor- begin;
      out[N]= malloc(L+1);
      memcpy(out[N], begin, L);
      out[N][L]= 0;
   }
   N++;
   goto next_token;
fail:
   *err= begin-inp;
   return 0;
}
/* END FILE cmdline.c */
/* CONTINUE FILE main.c */
/* BEGIN FILE critbit.c */
#include <sys/types.h>
/* BEGIN FILE critbit.h */
#ifndef CRITBIT_H_INCLUDED
#define CRITBIT_H_INCLUDED
#ifdef __cplusplus
extern "C" {
#endif
typedef struct {
  void *root;
} strmap_t;
strmap_t *strmap_new();
void *strmap_find(strmap_t *map, char *key);
void *strmap_insert(strmap_t *map, char *key,void *data);
void *strmap_remove(strmap_t *map, char *key);
void *strmap_destroy(strmap_t *map);
#ifdef __cplusplus
};
#endif
#endif
/* END FILE critbit.h */
/* CONTINUE FILE critbit.c */
typedef struct
{
  void* child[2];
  unsigned internal:1, byte:23, mask:8;
} strmap_node_t;
static inline int strmap_dir
    (uint8_t *key,uint32_t len_key,strmap_node_t *node)
{
  uint8_t c;
  if (node->byte<len_key) c= key[node->byte];
                     else c= 0;
  return (1+(node->mask|c))>>8;
}
void* strmap_find(strmap_t *map,char *skey)
{
  uint8_t *key; unsigned int len_key;
  strmap_node_t *node;
  node= map->root;
  if (!node) return 0;
  key= (void*) skey;
  len_key= strlen(skey);
  while(node->internal)
     node= node->child[strmap_dir(key,len_key,node)];
  if (strcmp(skey, node->child[0])) return 0;
                               else return node->child[1];
}
static inline uint32_t strmap_make_mask
    (uint32_t A, uint32_t B)
{
  uint8_t diff;
  diff= A ^ B;
  diff|= diff>>1;
  diff|= diff>>2;
  diff|= diff>>4;
  return (diff&~(diff>>1))^255;
}
static inline strmap_node_t* strmap_leaf_node
    (uint8_t *key, void *data)
{
  strmap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 0;
  N->child[0]= key;
  N->child[1]= data;
  return N;
}
static inline strmap_node_t* strmap_internal_node
    (uint32_t byte, uint32_t mask)
{
  strmap_node_t *N;
  N= malloc(sizeof(*N));
  N->internal= 1;
  N->byte= byte;
  N->mask= mask;
  return N;
}
void* strmap_insert(strmap_t* map, char* skey, void *data)
{
  uint8_t *key; uint32_t len_key;
  strmap_node_t *node_elt;
  key= (void*) skey;
  if (!map->root)
  {
     map->root= strmap_leaf_node(key,data);
     return 0;
  }
  len_key= strlen(skey);
  node_elt= map->root;
  while(node_elt->internal)
    node_elt= node_elt->child[strmap_dir(key,len_key,node_elt)];
  uint32_t crit_byte;
  uint8_t  mask;
  uint8_t  *elt;
  int dir_elt;
  strmap_node_t *node_key;
  elt= node_elt->child[0];
  for(crit_byte=0;
      crit_byte<=len_key && elt[crit_byte]==key[crit_byte];
      crit_byte++)
          ;
  if (crit_byte>len_key) return node_elt->child[1];
  mask= strmap_make_mask(elt[crit_byte],key[crit_byte]);
  dir_elt= (1+(mask|elt[crit_byte]))>>8;
  node_key= strmap_internal_node(crit_byte,mask);
  node_key->child[1-dir_elt]= strmap_leaf_node(key,data);
  strmap_node_t *parent;
  strmap_node_t *node;
  parent= 0;
  node= map->root;
  while(node->internal)
  {
     if (node->byte > crit_byte) break;
     if (node->byte==crit_byte && node->mask > mask) break;
     parent= node;
     node= node->child[strmap_dir(key,len_key,node)];
  }
  node_key->child[dir_elt]= node;
  if (parent) { if (parent->child[0]==node) parent->child[0]= node_key;
                                    else    parent->child[1]= node_key; }
  else map->root= node_key;
  return 0;
}
void *strmap_remove(strmap_t* map, char *skey)
{
  uint8_t *key; uint32_t len_key;
  void *retv;
  if (!map->root) return 0;
  key= (void*) skey;
  len_key= strlen(skey);
  strmap_node_t *node;
  strmap_node_t *sibling;
  strmap_node_t *parent;
  strmap_node_t *gparent;
  node= map->root;
  gparent= 0;
  sibling= 0;
  parent= 0;
  while(node->internal)
  {
    int dir;
    dir= strmap_dir(key, len_key, node);
    gparent= parent;
    parent= node;
    sibling= node->child[1-dir];
    node= node->child[dir];
  }
  if (strcmp(node->child[0],skey)) return 0;
  retv= node->child[1];
  free(node);
  if (!gparent)
  {
     map->root= sibling;
     if (parent) free(parent);
     return retv;
  }
  if (gparent->child[0]==parent)
    gparent->child[0]= sibling;
  else
    gparent->child[1]= sibling;
  free(parent);
  return retv;
}
void *strmap_destroy(strmap_t *map)
{
  strmap_node_t *node;
  strmap_node_t *parent;
  strmap_node_t *gparent;
  strmap_node_t *sibling;
  void *retv;
  if (!map->root) return 0;
  parent= 0;
  gparent= 0;
  sibling= 0;
  node= map->root;
  while(node->internal)
  {
    gparent= parent;
    parent= node;
    sibling= node->child[1];
    node= node->child[0];
  }
  retv= node->child[1];
  free(node);
  if (!gparent)
  {
    map->root= sibling;
    if (parent) free(parent);
    return retv;
  }
  if (gparent->child[0]==parent)
    gparent->child[0]= sibling;
  else
    gparent->child[1]= sibling;
  free(parent);
  return retv;
}
strmap_t  *strmap_new()
{
  strmap_t *s;
  s= malloc(sizeof(*s));
  s->root= 0;
  return s;
}
/* END FILE critbit.c */
/* CONTINUE FILE main.c */
/* BEGIN FILE fnmanip.c */
/* This function does the following transformations:

    A/B/../C    -> A/C
    /./         -> /
    //          -> /
    ./A         -> A
*/
void fnmani_normalize(char *name)
{
   int len;
   int i,j;
   char *F,*G,*psearch;
   len= strlen(name);
again:
   if (name[0]==0) return ;
   if (name[0]=='.' && name[1]=='/') name[0]= name[1]= 0;
   else if ((F= strstr(name,"//")))  F[0]= 0;
   else if ((F= strstr(name,"/./"))) F[0]= F[1]= 0;
   else goto parent;
   goto compact;
parent:
   psearch= name;
next_parent:
   F= strstr(psearch,"/../");
   if (!F) return ;
   if (F==name) { F[0]= F[1]= F[2]= 0; goto compact; }
   G= F-1;
   while(G>=name && G[0]!='/') G--;
   if (G<name || G[0]!='/') { psearch= F+1; goto next_parent; }
   if (G[1]=='.' && G[2]=='.' && G+3==F) { psearch= F+1; goto next_parent; }
   while(G<F+3) { G[0]= 0; G++; }
compact:
   j= 0;
   for(i=0;i<len;i++) if (name[i]) name[j++]= name[i];
   name[j]= 0; len= j;
   goto again;
}
/* returns the directory of the path. ignores any trailing / characters */
char *fnmani_dir(char *path)
{
   char* E;
   path= strdup(path);
   fnmani_normalize(path);
   for(E=path+strlen(path)-2;E>=path;E--)
      if (E[0]=='/') break;
   if (E>path) { E[0]= 0; return path; }
   free(path);
   if (E<path) return strdup("");
   return strdup("/");
}
/* This function returns a sibling of the file referenced by 'ref'.
   Both ref and sib can contain directory names, here's how everything
   is done:

   let be:
        D(ref)   directory part of ref
        D(sib)   directory part of sib

   if D(sib) is absolute, return sib
   return D(ref) concat sib

   The resulting string is normalized, by processing stuff such as
   /../ etc. */
char *fnmani_sibling(char *ref, char *sib)
{
   char *Dref,*Dsib;
   char *R;
   Dsib= fnmani_dir(sib);
   if (Dsib[0]=='/') { free(Dsib); return strdup(sib); }
   free(Dsib);
   Dref= fnmani_dir(ref);
   if (Dref[0]==0) { free(Dref); return strdup(sib); }
   R= malloc(strlen(Dref)+1+strlen(sib)+1);
   sprintf(R, "%s/%s", Dref, sib);
   free(Dref);
   fnmani_normalize(R);
   return R;
}
/* We are assuming that 'parent' is a directory name. There is only
   one special case here, parent can be completely empty. This means
   that  it was obtained from dir() and there was no directory part
   in the argument to dir(). In this case, we simply return the child
   name directly. */
char *fnmani_child(char *parent, char *name)
{
   char *K;
   if (parent[0]==0) return strdup(name);
   K= malloc(strlen(parent)+1 +strlen(name)+1);
   sprintf(K, "%s/%s", parent, name);
   fnmani_normalize(K);
   return K;
}
/* END FILE fnmanip.c */
/* CONTINUE FILE main.c */
/* BEGIN FILE input.c */
strmap_t *defines, *noprot,*soprot;
char **files, **paths;
char *output_file_name;
void **ptrbuffer_collect(obuffer_t *B)
{
  size_t L;
  void **R;
  B->head= 0; L= 0;
  while(!obuffer_empty(B)) { obuffer_ptr(B); L++; }
  R= (void**) malloc(sizeof(void*)*(L+1));
  B->head= 0; L= 0;
  while(!obuffer_empty(B)) { R[L]= obuffer_ptr(B); L++; }
  R[L]= NULL;
  obuffer_free(B);
  return R;
}
char *strip_quotes(char *K)
{
  if (K[0]!='"') return K;
  memmove(K,K+1, strlen(K));
  K[strlen(K)-1]= 0;
  return K;
}
void check_args(char **cmd, int narg, char *fn, int lno)
{
  int i;
  for(i=1;i<=narg;i++)
    if (!cmd[i] || !cmd[i][0])
       exit(fprintf(stderr,"%s:%d: command %s needs %d non-empty argument%s\n",
                    fn, lno, narg, narg==1? "":"s"));
}
int read_input_file(char *fn)
{
  ifile_t *f;
  uint8_t *line; size_t len; int lno;
  char *a1,*a2;
  obuffer_t *Bpaths, *Bfiles;
  char **cmd; int dummy; int i;
  cmd= NULL;
  defines= strmap_new(); noprot= strmap_new(); soprot= strmap_new();
  Bpaths= obuffer_new(); Bfiles= obuffer_new();
  if (ifile_open(&f,fn))
     exit(fprintf(stderr,"error reading input file: %s\n", ifile_error(f)));
  line= NULL;
again:
  if (line) free(line);
  if (cmd) { if (cmd[0]) free(cmd[0]); free(cmd); }
  if (ifile_get_line(f,&line, &len, 1)) goto done;
  lno= ifile_line_number(f);
  cmd= parse_cmd_line(line, &dummy);
  if (dummy!=-1) exit(fprintf(stderr,"%s:%d: syntax error\n", fn, lno));
  if (!cmd || !cmd[0]) goto again;
  if (!strcmp(cmd[0],"define")) goto do_define;
  if (!strcmp(cmd[0], "file")) goto do_file;
  if (!strcmp(cmd[0], "noprot")) goto do_noprot;
  if (!strcmp(cmd[0], "search")) goto do_search;
  if (!strcmp(cmd[0], "output")) goto do_output;
  exit(fprintf(stderr,"%s:%d: unknown command %s\n",
                              fn, lno, cmd[0]));
done:
  files= (char**) ptrbuffer_collect(Bfiles);
  paths= (char**) ptrbuffer_collect(Bpaths);
  for(i=0;files[i];i++) files[i]= fnmani_sibling(fn, files[i]);
  for(i=0;paths[i];i++) paths[i]= fnmani_sibling(fn, paths[i]);
  ifile_close(f);
  return ;
do_define:
  check_args(cmd, 2, fn, lno);
  a1= strip_quotes(cmd[1]);
  a2= cmd[2];
  if (strmap_find(defines, a1))
    exit(fprintf(stderr,"%s:%d: multiple definitions for %s\n", fn, lno, a1));
  strmap_insert(defines, a1, a2);
  goto again;
do_noprot:
  check_args(cmd, 1, fn, lno);
  a1= cmd[1];
  if (a1[0]=='<') {
    strdel(a1,0,0);
    strdel(a1,-1,-1);
    if (!strmap_find(soprot,a1)) strmap_insert(soprot,a1,a1);
  } else {
    a1= strip_quotes(a1);
    if (!strmap_find(noprot, a1)) strmap_insert(noprot, a1, a1);
  }
  goto again;
do_file:
  check_args(cmd, 1, fn, lno);
  obuffer_put(Bfiles,"p",cmd[1]);
  goto again;
do_search:
  check_args(cmd, 1, fn, lno);
  obuffer_put(Bpaths, "p", cmd[1]);
  goto again;
do_output:
  check_args(cmd,1, fn, lno);
  output_file_name= strip_quotes(cmd[1]);
  goto again;
}
/* END FILE input.c */
/* CONTINUE FILE main.c */
FILE *output;
strmap_t *included_headers,*included_files;
/* BEGIN FILE include.c */
int get_include_params(uint8_t *buffer,size_t len,
                       uint8_t **Rname, int *type)
{
   size_t i,j,k;
   uint8_t *name,S;
   const char *clude= "include";
   for(i=0;i<len;i++) if (buffer[i]!=' ') break;
   if (buffer[i]!='#') return 1;
   for(i=i+1;i<len;i++) if (buffer[i]!=' ') break;
   if (i==len) return 1;
   for(j=0;clude[j] && i<len;i++, j++)
      if (buffer[i]!=(uint8_t) clude[j]) return 1;
   if (clude[j]) return 1;
   for(;i<len;i++) if (buffer[i]!=' ') break;
   if (i==len) return 1;
   if (buffer[i]=='<') { *type= 'S'; S= '>'; goto quoted; }
   if (buffer[i]=='"') { *type= 'L'; S= '"'; goto quoted; }
// since we come here after we have eliminated all trailing
// spaces, we can simply assume that everything between i and
// j are relevant.
   *type= 'V';
   i--;
   j= len;
   goto collect;
quoted:
   for(j=i+1;j<len;j++) if (buffer[j]==S) break;
   if (j==len) return 1;
collect:
   name= malloc(j-i);
   for(k=i+1;k<j;k++) name[k-i-1]= buffer[k];
   name[k-i-1]= 0;
   *Rname= name;
   return 0;
}
int lookup_variable(uint8_t **Rname, int *type)
{
  char *value;
  int i;
  char S;
  value= strmap_find(defines, *Rname);
  if (!value) return 1;
  value= strdup(value);
  if (value[0]=='<') { *type= 'S'; S= '>'; }
  else { *type='L'; S='"'; }
  memmove(value, value+1, strlen(value));
  for(i=0;value[i];i++) if (value[i]==S) { value[i]= 0; break; }
  *Rname= value;
  return 0;
}
void process_file(ifile_t *f,char *path);
void include_file(char *name, char *parent_path, int parent_lno)
{
   int i;
   char *path;
   ifile_t *f;
   path= fnmani_sibling(parent_path, name);
   if (!ifile_open(&f,path)) goto done;
   ifile_close(f);
   for(i=0;paths[i];i++)
   {
     path= fnmani_child(paths[i], name);
     if (!ifile_open(&f,path)) goto done;
     ifile_close(f);
   }
   exit(fprintf(stderr,"%s:%d: can't find file '%s'\n",
                      parent_path, parent_lno, name));
done:
   process_file(f, path);
}
void process_file(ifile_t *f,char *path)
{
   uint8_t *buffer; size_t len; int lno;
   uint8_t *include_file_name; int include_type;
   path= strdup(path);
   fnmani_normalize(path);
   if (strmap_find(included_files,path) && !strmap_find(noprot,path))
      return ;
   strmap_insert(included_files, path, path);
   fprintf(output,"/* BEGIN FILE %s */\n", path);
again:
   if (get_combined_line(f, &buffer, &len, &lno)) goto done;
   remove_comments(f, &buffer, &len);
   remove_trailing_spaces(buffer, &len);
   if (!buffer[0]) goto free_and_next;
   if (get_include_params(buffer,len, &include_file_name, &include_type))
      goto print_and_next;
   if (include_type=='V' && lookup_variable(&include_file_name, &include_type))
      goto print_and_next;
   if (include_type=='S')
   {
     if (!strmap_find(included_headers,include_file_name))
     {
        strmap_insert(included_headers, include_file_name, include_file_name);
        goto print_hdr_and_next;
     }
     if (strmap_find(soprot, include_file_name)) goto print_hdr_and_next;
     goto free_and_next;
   }
// now, include type is 'L'. we already modified 'V' and handled 'S' above.
   include_file(include_file_name, path, lno);
   fprintf(output,"/* CONTINUE FILE %s */\n", path);
   goto free_and_next;
print_and_next:
   fprintf(output,"%s\n", buffer);
free_and_next:
   free(buffer);
   goto again;
print_hdr_and_next:
   fprintf(output,"#include <%s>\n",include_file_name);
   goto free_and_next;
done:
   ifile_close(f);
   fprintf(output,"/* END FILE %s */\n", path);
}
/* END FILE include.c */
/* CONTINUE FILE main.c */
int main(int argc,char **argv)
{
   int i;
   if (argc!=2)
   {
 fprintf(stderr,
"C amalgamation tool.\n\n"
"THE BEER-WARE LICENSE (Revision 42):\n"
"<cinarus@yahoo.com> wrote this file. As long as you retain this notice you\n"
"can do whatever you want with this stuff. If we meet some day, and you think\n"
"this stuff is worth it, you can buy me a beer in return.\n"
"\n"
"I don't guarantee that this program will work at all or that it won't\n"
"delete all data on your disk. Use it at your own risk.\n"
"\nusage: %s input-file-name\n"
"\nThe input file consists of line-based commands.\n"
"   file file-name\n"
"       Adds the given file to the amalgamation.\n"
"   noprot file-name\n"
"       Normally, this program makes sure that each (#include)d file is \n"
"       inserted into the amalgamation only once. This command lifts such\n"
"       protection for the given file. If you want to do the same thing \n"
"       for system includes, simply enclose the file name in a <> pair.\n"
"   define var file-name\n"
"       This is used for stuff like #include SOME_HEADER. Here, 'var'\n"
"       can be a macro name or any other expression. If you use an \n"
"       expression which contains spaces, you can quote the expression\n"
"       using double quotes. In any case, var should look like exactly\n"
"       the way it does where it is (#include)d from (excepting the quotes,\n"
"       obviously).\n"
"   output file-name\n"
"       Defines the name of the output file. The default is amalgam-output.c\n"
"   search dir-name\n"
"       Adds dir-name to the list of directories to be searched when looking\n"
"       for a file because of an #include \"foo.h\" directive. For such\n"
"       directives, the file is searched for in the directory of the\n"
"       including file, after that it's searched for in the directories\n"
"       given by search commands. The order in which these commands are given\n"
"       is significant, the same order is used during the search.\n"
"   # comment\n"
"       Anything following a # character on a line is ignored.\n"
"\nIn all cases, file names can be enclosed in double quotes. However, no\n"
"escape encoding can be used in them. ie. they are only useful for embedding\n"
"spaces in the file names, which is a pretty bad idea if you ask me.\n"
"\nFile names are relative to the path of the file in which they are \n"
"specified. For example a source file path is taken relative to the path\n"
"of the input file. Similarly, a header file path is relative to the path\n"
"of the source file including it.\n"
"\n",
argv[0]);
       exit(1);
   }
   included_headers= strmap_new();
   included_files= strmap_new();
   read_input_file(argv[1]);
   if (!output_file_name) output_file_name= "amalgam-output.c";
   output= fopen(output_file_name,"w");
   comment_out= output;
   fprintf(output,
               "/* This file is an amalgamation of several source files.\n"
               "   Please look for BEGIN FILE,END FILE and CONTINUE FILE\n"
               "   strings to browse contents of individual files contained\n"
               "   here. */\n");
   for(i=0;files[i];i++)
   {
     ifile_t *f;
     if (ifile_open(&f,files[i]))
      exit(fprintf(stderr,"input file %s: %s\n", argv[1], ifile_error(f)));
     process_file(f,files[i]);
   }
   fclose(output);
   return 0;
}
/* END FILE main.c */
