
int get_include_params(uint8_t *buffer,size_t len, 
                       uint8_t **Rname, int *type)
{
   size_t i,j,k;
   uint8_t *name,S;
   const char *clude= "include";

   for(i=0;i<len;i++) if (buffer[i]!=' ') break;
   if (buffer[i]!='#') return 1;
   for(i=i+1;i<len;i++) if (buffer[i]!=' ') break;
   if (i==len) return 1;

   for(j=0;clude[j] && i<len;i++, j++)
      if (buffer[i]!=(uint8_t) clude[j]) return 1;
   if (clude[j]) return 1;

   for(;i<len;i++) if (buffer[i]!=' ') break;
   if (i==len) return 1;

   if (buffer[i]=='<') { *type= 'S'; S= '>'; goto quoted; }
   if (buffer[i]=='"') { *type= 'L'; S= '"'; goto quoted; }

   // since we come here after we have eliminated all trailing
   // spaces, we can simply assume that everything between i and
   // j are relevant.
   *type= 'V';
   i--;
   j= len;
   goto collect;

quoted:
   for(j=i+1;j<len;j++) if (buffer[j]==S) break;
   if (j==len) return 1;
collect:
   name= malloc(j-i);
   for(k=i+1;k<j;k++) name[k-i-1]= buffer[k]; 
   name[k-i-1]= 0;
   *Rname= name;
   return 0;
}

int lookup_variable(uint8_t **Rname, int *type)
{
  char *value;
  int i;
  char S;

  value= strmap_find(defines, *Rname);
  if (!value) return 1;

  value= strdup(value);
  if (value[0]=='<') { *type= 'S'; S= '>'; }
  else { *type='L'; S='"'; }

  memmove(value, value+1, strlen(value));
  for(i=0;value[i];i++) if (value[i]==S) { value[i]= 0; break; }

  *Rname= value;
  return 0;
}


void process_file(ifile_t *f,char *path);

void include_file(char *name, char *parent_path, int parent_lno)
{
   int i;
   char *path;
   ifile_t *f;

   path= fnmani_sibling(parent_path, name);
   
   if (!ifile_open(&f,path)) goto done;
   ifile_close(f);
   for(i=0;paths[i];i++)
   {
     path= fnmani_child(paths[i], name);
     if (!ifile_open(&f,path)) goto done;
     ifile_close(f);
   }
   exit(fprintf(stderr,"%s:%d: can't find file '%s'\n", 
                      parent_path, parent_lno, name));
done:
   process_file(f, path);
}


void process_file(ifile_t *f,char *path)
{
   uint8_t *buffer; size_t len; int lno;
   uint8_t *include_file_name; int include_type;

   path= strdup(path);
   fnmani_normalize(path);
   if (strmap_find(included_files,path) && !strmap_find(noprot,path))
      return ;

   strmap_insert(included_files, path, path);

   fprintf(output,"/* BEGIN FILE %s */\n", path);
again:
   if (get_combined_line(f, &buffer, &len, &lno)) goto done;

   remove_comments(f, &buffer, &len);
   remove_trailing_spaces(buffer, &len);
   if (!buffer[0]) goto free_and_next;

   if (get_include_params(buffer,len, &include_file_name, &include_type))
      goto print_and_next;

   if (include_type=='V' && lookup_variable(&include_file_name, &include_type))
      goto print_and_next;

   if (include_type=='S')
   {
     if (!strmap_find(included_headers,include_file_name))
     {
        strmap_insert(included_headers, include_file_name, include_file_name);
        goto print_hdr_and_next;
     }
     if (strmap_find(soprot, include_file_name)) goto print_hdr_and_next;
     goto free_and_next;
   }

   // now, include type is 'L'. we already modified 'V' and handled 'S' above.

   include_file(include_file_name, path, lno);
   fprintf(output,"/* CONTINUE FILE %s */\n", path);
   goto free_and_next;

print_and_next:
   fprintf(output,"%s\n", buffer);

free_and_next:
   free(buffer);
   goto again;
print_hdr_and_next:
   fprintf(output,"#include <%s>\n",include_file_name);
   goto free_and_next;
done:
   ifile_close(f);
   fprintf(output,"/* END FILE %s */\n", path);
}

