
strmap_t *defines, *noprot,*soprot;
char **files, **paths;
char *output_file_name;


void **ptrbuffer_collect(obuffer_t *B)
{
  size_t L;
  void **R;

  B->head= 0; L= 0;
  while(!obuffer_empty(B)) { obuffer_ptr(B); L++; }
  R= (void**) malloc(sizeof(void*)*(L+1));

  B->head= 0; L= 0;
  while(!obuffer_empty(B)) { R[L]= obuffer_ptr(B); L++; }
  R[L]= NULL;

  obuffer_free(B);
  return R;
}

char *strip_quotes(char *K)
{
  if (K[0]!='"') return K;
  memmove(K,K+1, strlen(K));
  K[strlen(K)-1]= 0;
  return K;
}

void check_args(char **cmd, int narg, char *fn, int lno)
{
  int i;
  for(i=1;i<=narg;i++)
    if (!cmd[i] || !cmd[i][0]) 
       exit(fprintf(stderr,"%s:%d: command %s needs %d non-empty argument%s\n",
                    fn, lno, narg, narg==1? "":"s"));
}


int read_input_file(char *fn)
{
  ifile_t *f;
  uint8_t *line; size_t len; int lno;
  char *a1,*a2;

  obuffer_t *Bpaths, *Bfiles;
  char **cmd; int dummy; int i;

  cmd= NULL;
  defines= strmap_new(); noprot= strmap_new(); soprot= strmap_new();
  Bpaths= obuffer_new(); Bfiles= obuffer_new();

  if (ifile_open(&f,fn))
     exit(fprintf(stderr,"error reading input file: %s\n", ifile_error(f)));
  
  line= NULL;

again:
  if (line) free(line);
  if (cmd) { if (cmd[0]) free(cmd[0]); free(cmd); }
  if (ifile_get_line(f,&line, &len, 1)) goto done;

  lno= ifile_line_number(f);
  cmd= parse_cmd_line(line, &dummy);
  if (dummy!=-1) exit(fprintf(stderr,"%s:%d: syntax error\n", fn, lno));
  if (!cmd || !cmd[0]) goto again;

  if (!strcmp(cmd[0],"define")) goto do_define; 
  if (!strcmp(cmd[0], "file")) goto do_file;
  if (!strcmp(cmd[0], "noprot")) goto do_noprot;
  if (!strcmp(cmd[0], "search")) goto do_search;
  if (!strcmp(cmd[0], "output")) goto do_output;
  exit(fprintf(stderr,"%s:%d: unknown command %s\n", 
                              fn, lno, cmd[0]));
done:
  files= (char**) ptrbuffer_collect(Bfiles);
  paths= (char**) ptrbuffer_collect(Bpaths);


  for(i=0;files[i];i++) files[i]= fnmani_sibling(fn, files[i]);
  for(i=0;paths[i];i++) paths[i]= fnmani_sibling(fn, paths[i]);

  ifile_close(f);
  return ;

do_define:
  check_args(cmd, 2, fn, lno);
  a1= strip_quotes(cmd[1]); 
  a2= cmd[2];
  if (strmap_find(defines, a1))
    exit(fprintf(stderr,"%s:%d: multiple definitions for %s\n", fn, lno, a1));
  strmap_insert(defines, a1, a2);
  goto again;
  
do_noprot:
  check_args(cmd, 1, fn, lno);
  a1= cmd[1];
  if (a1[0]=='<') {
    strdel(a1,0,0);
    strdel(a1,-1,-1);
    if (!strmap_find(soprot,a1)) strmap_insert(soprot,a1,a1);
  } else {
    a1= strip_quotes(a1);
    if (!strmap_find(noprot, a1)) strmap_insert(noprot, a1, a1);
  }
  goto again;

do_file:
  check_args(cmd, 1, fn, lno);
  obuffer_put(Bfiles,"p",cmd[1]);
  goto again;

do_search:
  check_args(cmd, 1, fn, lno);
  obuffer_put(Bpaths, "p", cmd[1]);
  goto again;
do_output:
  check_args(cmd,1, fn, lno);
  output_file_name= strip_quotes(cmd[1]);
  goto again;
}

