#include <stdarg.h>

#define OBUFFER_INITIAL_SIZE 128

typedef struct {
  uint8_t *data;
  size_t size, len, head;
} obuffer_t;

obuffer_t *obuffer_new()
{
   obuffer_t *V;
   V= malloc(sizeof(*V));
   V->size= OBUFFER_INITIAL_SIZE;
   V->len= V->head= 0;
   V->data= malloc(V->size);
   return V;
}

void obuffer_put(obuffer_t *V, const char *fmt, ...)
{
   va_list ap;
   int i;
   size_t size;

   size_t copy_sz; void *copy_p;
   int intv; double douv; void *ptrv;
   int copy;

   copy= 0;

again:
   size= 0;
   va_start(ap,fmt);
   for(i=0;fmt[i];i++)
   {
    switch(fmt[i])
    {
    case 'i': intv= va_arg(ap,int); copy_sz= sizeof(int);
                                    copy_p= &intv; break;
    case 'd': douv= va_arg(ap, double); copy_sz= sizeof(double);
                                        copy_p= &douv; break;
    case 'p': ptrv= va_arg(ap, void*); copy_sz= sizeof(void*);
                                       copy_p= &ptrv; break;
    case 'b': copy_sz= va_arg(ap, size_t); copy_p= va_arg(ap, void*); break;
    }
    if (copy) { memcpy(V->data+V->len, copy_p, copy_sz); V->len+= copy_sz; }
    else size+= copy_sz;
   }
   va_end(ap);
   
   if (copy) return ;
   copy++;
   if (V->len+size>V->size)
   {
     V->size+= V->size/2 + size;
     V->data= realloc(V->data, V->size);
   }
   goto again;
}

int obuffer_empty(obuffer_t *o)
{
   return o->head==o->len;
}


void obuffer_read(obuffer_t *o, void *dst, size_t sz)
{
   memcpy(dst, o->data+o->head, sz);
   o->head+= sz;
}

double obuffer_dou(obuffer_t *o)
  { double D; obuffer_read(o,&D,sizeof(D)); return D; } 

void* obuffer_ptr(obuffer_t *o)
  { void* D; obuffer_read(o,&D,sizeof(D)); return D; } 

int obuffer_int(obuffer_t *o)
  { int D; obuffer_read(o,&D,sizeof(D)); return D; } 

void obuffer_free(obuffer_t *o) { free(o->data); free(o); }


